//===================================================================================
// 	cdc.c
// 	USB CDC Class
//===================================================================================
#include <system.h> 
#include <memory.h>
#include "usb_defs.h" 
#include "spi.h"
#include "iocnt.h"
#include "cdc.h"

//for output debug string to SIO
//#define _SDEBUG_
#include "sdebug.h"

//===================================
//	Globals
//===================================
#define lobyte(x) (x & 0xff)
#define hibyte(x) (x >> 8)
#ifndef max
#define max(x,y)	(x)>(y) ? (x):(y)
#endif
#ifndef min
#define min(x,y)	(x)<(y) ? (x):(y)
#endif

#define VENDOR_ID		0x1232
#define PRODUCT_ID		0x0001
#define VENDOR_ID_H		0x12
#define VENDOR_ID_L		0x32
#define PRODUCT_ID_H	0x00
#define PRODUCT_ID_L	0x01

unsigned char USB_IsIdle;
unsigned char USB_status_device;
unsigned char USB_Curr_Config; 
unsigned char USB_address_pending;
unsigned char USB_dev_req;
unsigned char USWSTAT;
unsigned char USB_Interface [3];

//Error Data Counter
unsigned int USB_PID_ERR   = 0;	// error counters
unsigned int USB_CRC5_ERR  = 0;
unsigned int USB_CRC16_ERR = 0;
unsigned int USB_DFN8_ERR  = 0;
unsigned int USB_BTO_ERR   = 0;
unsigned int USB_WRT_ERR   = 0;
unsigned int USB_OWN_ERR   = 0;
unsigned int USB_BTS_ERR   = 0;

struct BufferDescriptorEntry BDT[64]@0x400;
unsigned char *UEPn=(unsigned char *)UEP0;	//for Array accsecc of uep0-15

struct RequestPacket  CurrentPacket;

// USB BufferSize
static const int UsbBufSize[] = {	
	8,		//EP0
	64,		//EP1
	64,		//EP2
	8};		//EP3

//	EP0 - Control Transfer
//  EP1 - Bulk In
//  EP2 - Bulk Out
//  EP3 - Interrupt
static const int UsbEpType[] = {	
	EP_CONTROL,			//EP0
	EP_BULK_IN,			//EP1
	EP_BULK_OUT,		//EP2
	EP_INT_IN	};		//EP3

//CDC data
volatile unsigned char CtrlLineBmp=0;	// bit 0=DTR,Bit 1=RTS
volatile unsigned char LineCoding[7]={0x80,0x25,0,0,0,0,8};	//9600,n,8,1

//===================================
//	ROM FIFO
//===================================
char *CFifoData=NULL;	//char data pointer of FIFO 
rom char *RFifoData;	//rom data pointer of FIFO 
char FifoSize=0;
char FifoPoint=0;

void FifoFlash()
{
	CFifoData=NULL;	//char data pointer of FIFO 
	RFifoData=NULL;	//rom data pointer of FIFO 
	FifoSize=0;
	FifoPoint=0;
}

void SetFifoData(char *fdat,char size)
{
	CFifoData=fdat;
	RFifoData=NULL;
	FifoSize=size;
	FifoPoint=0;
}

void SetFifoData(rom char *fdat,char size)
{
	RFifoData=fdat;
	CFifoData=NULL;
	FifoSize=size;
	FifoPoint=0;
}

char GetFifoRemain()
{
	return (FifoSize-FifoPoint);
}

int GetFifoData()
{
	int dat;
	if((FifoSize==0)||(FifoSize==FifoPoint)||((RFifoData==NULL)&&(CFifoData==NULL)))
	   return -1;
	if(RFifoData)
		dat=RFifoData[FifoPoint];
	else
		dat=CFifoData[FifoPoint];
	FifoPoint++;
	return dat;
}

//===================================
//	Descriptors
//===================================

//===================================
//	String descriptors
//===================================
rom char *String0 = {4,STRING,9,4};
rom char *String1 = {24, STRING,'S',0,'o',0,'l',0,'i',0,'t',0,'o',0,'n',0,'W',0,'a',0,'v',0,'e',0};
rom char *String2 = {56,STRING,'P',0,'I',0,'C',0,'1',0,'8',0,'F',0,'4',0,'4',0,'5',0,'0',0,' ',0,
		'C',0,'D',0,'C',0,' ',0,'S',0,'a',0,'m',0,'p',0,'l',0,'e',0,' ',0, 'D',0,'e',0,'m',0,'o',0,'.',0};

//===================================
//	Device descriptor
//===================================
rom char *DeviceDescriptor= {
	0x12, 		    // 18 bytes long 
	DEVICE,		    // descriptor type
	0x10, 0x01,	    // USB specification release (1.10)
	CDC_DEVICE,	    // class code
	0x00, 		    // subclass code
	0x00, 		    // protocol code
	EP0_BUFF_SIZE,	// Max packet size for EP0
	VENDOR_ID_L,VENDOR_ID_H,
	PRODUCT_ID_L,PRODUCT_ID_H,
	0x00,0x00, 	    // device release number 2.00
	0x01, 		    // index to string that describes vendor
	0x02, 		    // index to string that describes product
	0x00, 		    // index to string that describes serial number (none)
	0x01  		    // number of possible configurations
};
//===================================
//	Config descriptor
//	EP0 - Control Transfer
//  EP1 - Bulk In
//  EP2 - Bulk Out
//  EP3 - Interrupt
//===================================
#define CONFIG_DESC_SIZE 67
rom char *ConfigDescriptor  = {
	/* Configuration Descriptor */
	0x09,						// Size of Configuration Descriptor
	0x02,						// CONFIGURATION descriptor type
	CONFIG_DESC_SIZE,			// Total length of data for this cfg(L)
	0,							// Total length of data for this cfg(H)
	2,							// Number of interfaces in this cfg
	1,							// Index value of this configuration
	0,							// Configuration string index
	0xA0,						// configuration attributes , remote wakeup
	0x32,						// current consumption in 2mA units (100 mA) 

	/* Interface Descriptor */
	0x09,					// Size of this descriptor in bytes
	0x04,				    // INTERFACE descriptor type
	0,						// Interface Number
	0,						// Alternate Setting Number
	1,						// Number of endpoints in this intf
	COMM_INTF,				// Class code
	ABSTRACT_CONTROL_MODEL,	// Subclass code
	V25TER,					// Protocol code
	0,						// Interface string index

	/* CDC Class-Specific Descriptors */
	0x05,0x24,0x00,0x10,0x01,	//USB_CDC_HEADER_FN_DSC
	0x04,0x24,0x02,0x02,		//USB_CDC_ACM_FN_DSC
	0x05,0x24,0x06,0x00,0x01,	//USB_CDC_UNION_FN_DSC
	0x05,0x24,0x01,0x00,0x01,	//USB_CDC_CALL_MGT_FN_DSC

	/* Endpoint Descriptor */
	0x07,
	ENDPOINT,
	0x83,		// endpoint characteristics (IN endpoint)
	0x03,		// endpoint attributes  (interrupt)
	0x08, 0x00,	// max packet size (8)
	0x02,		//polling interval=2 msec


	/* Interface Descriptor */
	0x09,					// Size of this descriptor in bytes
	INTERFACE,				// INTERFACE descriptor type
	1,						// Interface Number
	0,						// Alternate Setting Number
	2,						// Number of endpoints in this intf
	DATA_INTF,				// Class code
	0,						// Subclass code
	NO_PROTOCOL,			// Protocol code
	0,						// Interface string index

	/* Endpoint Descriptors */
	0x07,
	ENDPOINT,
	0x81,		// endpoint characteristics (IN endpoint)
	0x02,		// endpoint attributes  (Bulk)
	64, 0x00,	// max packet size (64)
	0x00,
	0x07,
	ENDPOINT,	
	0x02,		// endpoint characteristics (out endpoint)
	0x02,		// endpoint attributes  (Bulk)
	64, 0x00,	// max packet size (64)
	0x00
};
//===================================
//	Private Prototypes
//===================================
//USB Service Functions
void USBStall ();		//Stall
void Count_Error ();	//Error
void USBSleep ();		//Sleep
void SendEndpoint0();		//Control Transfer Functions
void ReceiveEndpoint0();	//Control Transfer Functions
void TokenDone();		//Token Done
void InTokenDone(char ep);		//Called at IN Token done (EP>0)
void OutTokenDone(char ep);		//Called at OUT Token done (EP>0)
void StdRequest();		//Setup Functions (Standard)
void ClassRequest();	//Setup Functions (CDC)
//USB-SPI Data Interface
void USB_SPI_DataExchange();

//===================================
//	USB Initialize Functions
//===================================
void InitUSB ()
{
	char i,eptype,ep;
	int bufadr,bufsize;

	//Initiaize BDT
	bufadr=0x500;
	for(i=0;i<=6;i++){
		ep=i/2;
		bufsize=UsbBufSize[ep];
		eptype=UsbEpType[ep];
		if((eptype==EP_CONTROL)||((eptype%2)==(i%2))){
			BDT[i].address=bufadr;
			BDT[i].bytes=bufsize;
			bufadr+=bufsize;
		}else{
			BDT[i].address=0;
			BDT[i].bytes=0;
			BDT[i].EPStat=0;
		}
	}

	ucfg=  00010100b; //use onchip receiver, internal pullup resistors, full
	//            speed and no pingpong buffers
	uie = 00000001b;	// enable the Reset interrupt ONLY!!
	uir         = 0;	// clear all USB interrupt flags
	pie2= 00100000b ; //enable usb interrupts
	ucon = 01000000b;	// reset pingpong buffers
	ucon = 00001000b;	// unlock pingpong buffers and enable usb
	USB_Curr_Config   = 0;
	USB_status_device = 1;
	for (i=0;i<3;i++)
		USB_Interface [i] =0;
	USB_dev_req       = NULL;
	// clear all error counters
	USB_PID_ERR       = 0;
	USB_CRC5_ERR      = 0;
	USB_CRC16_ERR     = 0;
	USB_DFN8_ERR      = 0;
	USB_BTO_ERR       = 0;
	USB_BTS_ERR       = 0;
	intcon  = intcon | 11000000b;	// Enable GIE & PEIE
	dputs("USB Initialized.");
}
//===================================
//	USB Reset Functions
//===================================
void USBReset ()
{
	portb = 1;		// Lite led 0 to indicate Reset status
	USB_Curr_Config = 0;
	USB_IsIdle   = 0;
	bTOK_DONE=0; // clear TOK_DNE bit in uir 4 times to 
	bTOK_DONE=0; // clear out the USTAT FIFO.  See Microchip 	
	bTOK_DONE=0; // 18F2455 data sheet.
	bTOK_DONE=0;
	//EP0 - ControlTransfer
	BDT[EP0OUT].bytes=CTRL_PKTSZ;
	BDT[EP0OUT].EPStat=BDT_UOWN | BDT_DTSEN;
	BDT[EP0IN].EPStat=BDT_DTSEN;
	//EP1 - Bulk in Transfer
	BDT[EP1IN].EPStat=BDT_DTS | BDT_DTSEN;
	//EP2 - Bulk out Transfer
	BDT[EP2OUT].EPStat=BDT_UOWN | BDT_DTSEN;
	BDT[EP2OUT].bytes=BULK_PKTSZ;
	//EP3 - Interrupt Transfer
	BDT[EP3IN].EPStat=BDT_DTSEN;
	uaddr    = 0; // clear address
	uir      = 0; // clear USB interrupt flag
	uie  = 00000101b; // enable reset and activity interrupt
	uep0     = ENDPT_CONTROL;
	uep1		= ENDPT_IN_ONLY;
	uep2		= ENDPT_OUT_ONLY;
	uep3		= ENDPT_IN_ONLY;
	USWSTAT=DEFAULT_STATE;
	USB_status_device = 1;
}
//===================================
//	USB Activity Functions
//===================================
void USBActivity ()
{
	bACTIVITY   = 0;
	bACTIVITY_E = 0;
	bSUSPND     = 0; //normal operation
	USB_IsIdle  = 0;
}
//===================================
//	USB Service Functions
//===================================
/* ********************************************************************* */
/* Branch off and service the USB interrupt flags	                     */
/* ********************************************************************* */
void UsbInt ()
{
	if (bTOK_DONE)
		TokenDone();
	if (bSTALL) {
#ifdef DEBUG_SERIAL						
		dputs("Unstall");
#endif		
		USBStall ();
	}
	if (bUERR) {
#ifdef DEBUG_SERIAL						
		dputs("Error");
#endif	
		Count_Error ();
	}
	if (bUIDLE) {
#ifdef DEBUG_SERIAL						
		dputs("Sleep");
#endif		
		USBSleep ();
	}
}
//===================================
//	USB Stall Functions
//===================================
void USBStall ()
{
	bSTALL = 0;
}
//===================================
//	USB ErrorCount Functions
//===================================
void Count_Error ()
{
	if (bPID_ERR && bPID_ERR_E) {
		++ USB_PID_ERR;
	}
	if (bCRC5 && bCRC5_E) {
		++ USB_CRC5_ERR;
	}
	if (bCRC16 && bCRC16_E) {
		++ USB_CRC16_ERR;
	}
	if (bDFN8 && bDFN8_E) {
		++ USB_DFN8_ERR;
	}
	if (bBTO_ERR && bBTO_ERR_E) {
		++ USB_BTO_ERR;
	}
	if (bBTS_ERR && bBTS_ERR_E) {
		++ USB_BTS_ERR;
	}
	ueir = 0; //Clear all USB error flag bits
	bUERR = 0; //Clear master USB error flag bit
}
//===================================
//	USB Sleep Functions
//===================================
void USBSleep ()
{
	bACTIVITY_E = 1; // enable activity interrupt
	bUIDLE      = 0; //disable idle interrupt
	bSUSPND    = 1;	 // put in suspend mode
	USB_IsIdle  = 1; // mark device as idle
}
//===================================
//	USB Process Request Functions
//===================================
void InTokenDone(char ep)
{
	//Add process here
}

void OutTokenDone(char ep)
{
	if(ep==2){
		//BulkOutɃf[^MꂽꍇASPIƒʐMs
		USB_SPI_DataExchange();
	}
}
void TokenDone()
{
	unsigned char  ReqType;
	unsigned char  PID;
	struct RequestPacket *pRPacket;
	char ep;
	unsigned char USB_USTAT;
	unsigned char ACTIVE_BUF;
	struct BufferDescriptorEntry BDTCopy; 

	USB_USTAT       = ustat;
	ACTIVE_BUF  =(USB_USTAT >>3)*2;						// EP*2 (IF OUTPUT)
	ACTIVE_BUF =(ACTIVE_BUF + ((USB_USTAT&0x4)>>2));	// =EP*2 + 1 (IF INPUT) 
	ep=(USB_USTAT >>3) & 0x0f;

	//save data in buffer descriptor table
	do{	
		BDTCopy.EPStat  = BDT[ACTIVE_BUF].EPStat;
	}while((BDTCopy.EPStat & 0x80) != 0);			//UOWN bit must be 0
	BDTCopy.bytes   = BDT[ACTIVE_BUF].bytes;
	BDTCopy.address = BDT[ACTIVE_BUF].address;
	bTOK_DONE        = 0;							//Clearing TRNIF (USTAT FIFO Advance)//]荞݂̃NAîƂUSTATFIFOXVj

	//Mg[NɂA𕪂
	PID = (BDTCopy.EPStat>>2) & 0x0F;
	switch(PID) {
	case PID_IN :	// TOKEN IN̏ꍇ
		if(ep==0){
			SendEndpoint0 ();
		}else{
			InTokenDone(ep);
		}
		break;
	case PID_OUT:	// TOKEN OUT̏ꍇ
		if(ep==0){
			ReceiveEndpoint0();
		}else{
			OutTokenDone(ep);
		}
		break;
	case PID_SETUP:	// TOKEN SETUP̏ꍇ
		//obt@̃|C^̏
		pRPacket =  (struct RequestPacket *)BDT[EP0OUT].address;
		CurrentPacket.bmRequestType = pRPacket->bmRequestType;
		CurrentPacket.bRequest      = pRPacket->bRequest;
		CurrentPacket.wValue        = pRPacket->wValue;
		CurrentPacket.wIndex        = pRPacket->wIndex;
		CurrentPacket.wLength       = pRPacket->wLength;
		bPID_ERR =0;	// Clear REQUEST ERROR  
		BDT[EP0OUT].bytes=CTRL_PKTSZ;
		BDT[EP0OUT].EPStat=BDT_UOWN | BDT_DTSEN;
		BDT[EP0IN].EPStat=BDT_DTSEN;
		bPKT_DIS    = 0;
		USB_dev_req = 0;
		ReqType=CurrentPacket.bmRequestType & 0x60 >> 5;
		switch(ReqType){
		case 0:
			dputs("Standerd Request");
			StdRequest();
			break;
		case 1:
			dputs("Class Request");
			ClassRequest();
			break;
		case 2:
			dputs("Vendor Request");
			//VendorRequest();
			break;
		default:
		}
		break;
	default:
		dputs("IRREGULAR TOKEN");
		break;
	}
}

//===================================
//	USB Setup Functions	/ Standard Reqest
//===================================
char GetDescriptor(char DescID,int len,char index)
{
	//DescID=Descriptor ID,len=SetLength,index=DescriptorIndex(for String Descriptor)
	int wLen;
	char retc=1;

	if (DescID == DEVICE) {
		//Device Descriptor
		wLen=min(DeviceDescriptor[0],len);
		SetFifoData(DeviceDescriptor,wLen);
	} else if (DescID == CONFIGURATION) {
		//ConfigurationDescriptor
		wLen=min(CONFIG_DESC_SIZE,len);
		SetFifoData(ConfigDescriptor,wLen);
	} else if (DescID == STRING) {
		//StringDescripter
		switch (index) {
		case 0:
			wLen=min(String0[0],len);
			SetFifoData(String0,wLen);
			break;
		case 1:
			wLen=min(String1[0],len);
			SetFifoData(String1,wLen);
			break;
		case 2:
			wLen=min(String2[0],len);
			SetFifoData(String2,wLen);
			break;
			// Additional string processing can be added here						
		default:
			retc=0;
		}
	}else{
		retc=0;
	}
	if(retc){
		USB_dev_req = GET_DESCRIPTOR;
		SendEndpoint0();
	}
	return retc;
}

void StdRequest()
{
	unsigned char  *OutBuffer;
	unsigned char  *UEPArray;
	unsigned char  DescriptorType;
	unsigned char  Endpoint;
	unsigned char  Interface;
	unsigned char  DescriptorID;
	unsigned char  wLen;
	bool beStall;
	char Target;

	beStall=true;
	Target=CurrentPacket.bmRequestType;
	switch (CurrentPacket.bRequest) {
	case CLEAR_FEATURE:	//0x01h
		if (CurrentPacket.wValue == 1) {
			switch(Target){
			case HOSTTODEVICE:
				if (CurrentPacket.wValue == 1) {
					USB_status_device &= 0xFD;
					Send_0Len_pkt;
					beStall=false;
				}
				break;
			case HOSTTOINTERFACE:
				//ClearFeature not support -- Stall
				break;
			case HOSTTOENDPOINT:
				UEPArray = (unsigned char *) &uep0;	//UEPArray[] -> UEPn[]
				Endpoint = CurrentPacket.wIndex & 0x0F;
				/* Only valid feature is 0 (Remote Wakeup) */
				if (CurrentPacket.wValue == 0){
					if (((USWSTAT & 0x03) == ADDRESS_STATE) && (Endpoint == 0)) {
						UEPArray [Endpoint] &= 0xFE;
						Send_0Len_pkt;
						beStall=false;
					} else if (((USWSTAT& 0x03) == CONFIG_STATE) && (Endpoint < 3)) {
						UEPArray [Endpoint] &= 0xFE;
						Send_0Len_pkt;
						beStall=false;
					}
				}
				break;
			}
		}
		break;
	case GET_CONFIGURATION:
		if(Target==DEVICETOHOST){
			OutBuffer =  (unsigned char *) BDT[EP0IN].address;
			OutBuffer [0] = USB_Curr_Config;
			BDT[EP0IN].bytes = 1;
			BDT[EP0IN].EPStat=BDT_UOWN | BDT_DTS | BDT_DTSEN;
			beStall=false;
		}
		break;
	case GET_DESCRIPTOR:
		if(Target==DEVICETOHOST){
			DescriptorID = (unsigned char) (CurrentPacket.wValue >> 8);
			if(GetDescriptor(DescriptorID,CurrentPacket.wLength,CurrentPacket.wValue & 0xff))
				beStall=false;
		}
		break;
	case GET_INTERFACE:
		if(Target==INTERFACETOHOST){
			Interface = (unsigned char)CurrentPacket.wIndex;
			if ((USWSTAT== CONFIG_STATE) && (Interface < NUM_INTERFACES)) {
				OutBuffer = (unsigned char *) BDT [EP0IN].address;
				OutBuffer[0] = USB_Interface [Interface];
				BDT[EP0IN].bytes=1;
				BDT[EP0IN].EPStat=BDT_UOWN | BDT_DTS | BDT_DTSEN;
				beStall=false;
			}
		}
		break;
	case GET_STATUS:
		switch(Target){
		case DEVICETOHOST:
			OutBuffer = (unsigned char *)BDT [EP0IN].address;
			OutBuffer[0] = USB_status_device;
			OutBuffer[1] = 0;
			BDT[EP0IN].bytes=2;
			BDT[EP0IN].EPStat=BDT_UOWN | BDT_DTS | BDT_DTSEN;
			beStall=false;
			break;
		case INTERFACETOHOST:
			OutBuffer = (unsigned char *) BDT [EP0IN].address;
			OutBuffer[1] = 0;
			BDT[EP0IN].bytes=2;
			Interface = (unsigned char)CurrentPacket.wIndex;
			if ((USWSTAT == ADDRESS_STATE) && (Interface == 0)) {
				OutBuffer[0] = USB_Interface [Interface];
				BDT[EP0IN].EPStat=BDT_UOWN | BDT_DTS | BDT_DTSEN;
			} else if ((USWSTAT == CONFIG_STATE) && (Interface < NUM_INTERFACES)) {
				OutBuffer[0] = USB_Interface [Interface];
				BDT[EP0IN].EPStat=BDT_UOWN | BDT_DTS | BDT_DTSEN;
			}
			beStall=false;
			break;
		case ENDPOINTTOHOST:
			UEPArray = (unsigned char *) &uep0;
			Endpoint = CurrentPacket.wIndex & 0x0F;
			OutBuffer=(unsigned char *)BDT[EP0IN].address;
			OutBuffer[1] = 0;
			BDT[EP0IN].bytes=2;
			if (Endpoint < 3) {
				OutBuffer[0] = UEPArray [Endpoint] & 0x01;
				BDT[EP0IN].EPStat=BDT_UOWN | BDT_DTS | BDT_DTSEN;
				beStall=false;
			}
			break;
		}
		break;
	case SET_ADDRESS:
		if(Target==HOSTTODEVICE){
			USB_address_pending =(unsigned char)CurrentPacket.wValue;
			if (USB_address_pending < 0x80) {
				Send_0Len_pkt;
				USB_dev_req = SET_ADDRESS;
				uie = 00001001b; //enable transaction complete and reset interrupts
				beStall=false;
			}
		}
		break;
	case SET_CONFIGURATION:
		if(Target==HOSTTODEVICE){
			if (CurrentPacket.wValue <= NUM_CONFIGURATIONS){
				USB_Curr_Config = (unsigned char)CurrentPacket.wValue;
			}
			if (CurrentPacket.wValue == 0) {
				USWSTAT= ADDRESS_STATE;
			} else {
				USWSTAT= CONFIG_STATE;
			}
			Send_0Len_pkt;
			uep3 = ENDPT_NON_CONTROL;
		}
		break;
#if 0
	case SET_DESCRIPTOR:
		break;
#endif
	case SET_FEATURE:
		if(Target==HOSTTODEVICE){
			if ((CurrentPacket.wValue & 0xff) == 1) {
				USB_status_device |= 0x02;	/* feature is remote wakeup */
				Send_0Len_pkt;
				beStall=false;
			}
		}else if(Target==HOSTTOENDPOINT){
			UEPArray = (unsigned char *) &uep0;
			Endpoint = CurrentPacket.wIndex & 0x0F;
			if (CurrentPacket.wValue == 0) {  /* Only valid feature is 0 (Remote Wakeup) */
				if (((USWSTAT& 0x03) == ADDRESS_STATE) && (Endpoint == 0)) {
					UEPArray [Endpoint] |= 1;
					Send_0Len_pkt;
					beStall=false;
				} else if (((USWSTAT& 0x03) == CONFIG_STATE) && (Endpoint < 3)) {
					UEPArray [Endpoint] |= 1;
					Send_0Len_pkt;
					beStall=false;
				}
			}
		}
		break;
	case SET_INTERFACE:
		if(Target==HOSTTOINTERFACE){
			if (USWSTAT == CONFIG_STATE) {
				Interface = (unsigned char)CurrentPacket.wIndex;
				USB_Interface [Interface] = (unsigned char)CurrentPacket.wValue;
				Send_0Len_pkt;
				beStall=false;
			}
		}
		break;
#if 0
	case SYNC_FRAME:
		break;
#endif
	default:
		beStall=true;
	}
	if(beStall)
		STALL_EP0;
}

//===================================
//	USB Setup Functions	/ CDC Class Reqest
//===================================
void ClassRequest()
{

	if(CurrentPacket.bmRequestType & 0x80){
		//In
		if(CurrentPacket.bRequest==GET_LINE_CODING){
			//GetLineCoding
			dputs("GetLineCoding");
			SetFifoData(LineCoding,7);
			SendEndpoint0();
		}
	}else{
		//out
		switch(CurrentPacket.bRequest){
		case SET_LINE_CODING:
			//SetLineCoding
			dputs("SetLineCoding");
			break;
		case SET_CONTROL_LINE_STATE:
			dputs("SetControlLineState");
			CtrlLineBmp=(unsigned char)CurrentPacket.wValue;
			SetLed(CtrlLineBmp & 0x01);	//SPIfoCXCSs̏ԂLEDɕ\
			SpiCS(CtrlLineBmp & 0x01);	//SPIfoCXCSsDTȐԂŐ
			break;
		default:
			break;
		}
		FifoFlash();
		SendEndpoint0();
	}
}
//===================================
//	USB Copy Descriptor Functions
//===================================
char iWriteEndpoint(char EndPoint,char *buff,char size)
{
	char *UsbBuffer;	/* pointer to the USB Buffers */
	char BufNo;
	char maxsize;
	char i;
	//BufNo=EndPoint*2+1;	//Point to MOde 0,BDT-IN 
	maxsize=UsbBufSize[EndPoint];
	BufNo=EP2BF(EndPoint,EP_IN);
	if(size>maxsize)
		size=maxsize;
	UsbBuffer =  (unsigned char *)BDT[BufNo].address;
	//memcpy(UsbBuffer,buff,size);	//For thread safe
	for(i=0;i<size;i++)
		UsbBuffer[i]=buff[i];
	BDT[BufNo].bytes=size;
	BDT[BufNo].EPStat &= BDT_DTS;		/* save only the DTS bit */
	BDT[BufNo].EPStat ^= BDT_DTS;		/* toggle data DTS bit */
	BDT[BufNo].EPStat |= BDT_UOWN | BDT_DTSEN;		/* set OWN and DTSEN bits */
	return size;
}
/*===================================================================
	SenEndpoin0()
	Send Fifo Data to Endpoint-0
===================================================================*/
void SendEndpoint0()
{
	char reqsize,wsize;
	char buff[EP0_BUFF_SIZE];

	reqsize=0;
	while(GetFifoRemain()){
		buff[reqsize]=GetFifoData();
		reqsize++;
		if(reqsize>=EP0_BUFF_SIZE)
			break;
	}
	wsize=iWriteEndpoint(0,buff,reqsize);
	if(GetFifoRemain()==0)
		USB_dev_req = 0;	/* Yes, clear the device reqest */
}

/*===================================================================
	ReceiveEndpoint0()
	Receive data from endpoint0
	(No function)
===================================================================*/
void ReceiveEndpoint0()
{
	//No Function
}

/*===================================================================
	char ReadEndpoint(char endpoint,char *buff,char size)
	endpoint:endpoint number (>0)
	buff:	read buffer
	size:	buffer size
	return : read bytes
===================================================================*/
char ReadEndpoint(char endpoint,char *buff,char size)
{
	char BdtNo;

	BdtNo=EP2BF(endpoint,EP_OUT);
	if((BDT[BdtNo].EPStat & 0x80) != 0)
		return 0;					//No data recieved

	//Copy Data 
	if(BDT[BdtNo].bytes<size)
		size=BDT[BdtNo].bytes;
	memcpy(buff,(char *)(BDT[BdtNo].address),size);

	BDT[BdtNo].bytes = UsbBufSize[endpoint];	//Reset counter for buffer size
	BDT[BdtNo].EPStat &= BDT_DTS;	//Clear except DTS bit
	BDT[BdtNo].EPStat ^= BDT_DTS;	//Toggle DTS bit
	BDT[BdtNo].EPStat |= BDT_UOWN | BDT_DTSEN;	//Set UOWN bit(semaphore) and DTSEN bit (Toggle sync continue)

	return size;
}

/*===================================================================
	char WriteEndpoint(char endpoint,char *buff,char size)
	endpoint:endpoint number (>0)
	buff:	write buffer
	size:	buffer size
	return : wite bytes
===================================================================*/
char WriteEndpoint(char endpoint,char *buff,char size)
{
	char BdtNo;
	char maxsize;
	maxsize=UsbBufSize[endpoint];

	BdtNo=EP2BF(endpoint,EP_IN);
	if((BDT[BdtNo].EPStat & 0x80) != 0)
		return 0;				
	if(size>maxsize)
		size=maxsize;
	memcpy((char *)(BDT[BdtNo].address),buff,size);
	BDT[BdtNo].bytes = size;
	BDT[BdtNo].EPStat &= BDT_DTS;
	BDT[BdtNo].EPStat ^= BDT_DTS;
	BDT[BdtNo].EPStat |= BDT_UOWN | BDT_DTSEN;
	return size;
}

//======================================================================
//	USB-SPI Data Interface FunctionsiSPI̒ʐMj
//======================================================================
void USB_SPI_DataExchange()
{
	char i;
	unsigned char buff[64];
	unsigned char len;

	len=BulkOut(buff,64);	//BulkOutɑMꂽf[^擾
							//len͎ۂɎMTCY
	dputs("Write Data Dump:");
	for(i=0;i<len;i++){
		dhex(buff[i]);
		buff[i]=SpiRW(buff[i]);	//f[^SPIɓ]AɎM
	}
	new_line();
	dputs("ReadData Dump:");
	for(i=2;i<len;i++){
		dhex(buff[i]);
	}
	new_line();
	BulkIn(buff,len);	//SPĨf[^BulkInɑM
}

